"""
A subpackage for simulating the interaction of light with transparent
or reflective materials.
"""

from ._snells_law import *
from . import profiles
from . import matrices
from ._layers import *
from ._materials import *
from ._multilayers import *



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from optika.materials._tests.test_materials import AbstractTestAbstractMaterial


@pytest.mark.parametrize(
    argnames="wavelength,result_expected",
    argvalues=[
        (1.0 * u.eV, 0),
        (2.0 * u.eV, 1),
        (2 * optika.sensors.energy_electron_hole, 2),
    ],
)
def test_quantum_yield_ideal(
    wavelength: u.Quantity | na.AbstractScalar, result_expected: na.AbstractScalar
):
    result = optika.sensors.quantum_yield_ideal(wavelength)
    assert np.all(result == result_expected)


@pytest.mark.parametrize(
    argnames="wavelength",
    argvalues=[
        304 * u.AA,
        na.linspace(100, 200, axis="wavelength", num=4) * u.AA,
    ],
)
@pytest.mark.parametrize(
    argnames="direction",
    argvalues=[
        None,
        na.Cartesian3dVectorArray(0, 0, 1),
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_oxide",
    argvalues=[
        10 * u.AA,
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_implant",
    argvalues=[
        1000 * u.AA,
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_substrate",
    argvalues=[
        1 * u.um,
    ],
)
@pytest.mark.parametrize(
    argnames="cce_backsurface",
    argvalues=[
        0.2,
        1,
    ],
)
def test_quantum_efficiency_effective(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: na.AbstractCartesian3dVectorArray,
    thickness_oxide: u.Quantity | na.AbstractScalar,
    thickness_implant: u.Quantity | na.AbstractScalar,
    thickness_substrate: u.Quantity | na.AbstractScalar,
    cce_backsurface: u.Quantity | na.AbstractScalar,
):
    result = optika.sensors.quantum_efficiency_effective(
        wavelength=wavelength,
        direction=direction,
        thickness_oxide=thickness_oxide,
        thickness_implant=thickness_implant,
        thickness_substrate=thickness_substrate,
        cce_backsurface=cce_backsurface,
    )
    assert np.all(result >= 0)
    assert np.all(result <= 1)


class AbstractTestAbstractImagingSensorMaterial(
    AbstractTestAbstractMaterial,
):
    pass


class AbstractTestAbstractCCDMaterial(
    AbstractTestAbstractImagingSensorMaterial,
):
    pass


class AbstractTestAbstractBackilluminatedCCDMaterial(
    AbstractTestAbstractCCDMaterial,
):
    def test_thickness_oxide(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
    ):
        result = a.thickness_oxide
        assert result >= 0 * u.mm

    def test_thickness_implant(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
    ):
        result = a.thickness_implant
        assert result >= 0 * u.mm

    def test_thickness_substrate(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
    ):
        result = a.thickness_substrate
        assert result >= 0 * u.mm

    def test_cce_backsurface(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
    ):
        result = a.cce_backsurface
        assert result >= 0

    @pytest.mark.parametrize(
        argnames="wavelength",
        argvalues=[
            500 * u.nm,
        ],
    )
    def test_quantum_yield_ideal(
        self,
        a: optika.sensors.AbstractBackilluminatedCCDMaterial,
        wavelength: u.Quantity | na.AbstractScalar,
    ):
        result = a.quantum_yield_ideal(wavelength)
        assert result >= 0


class AbstractTestAbstractStern1994BackilluminatedCCDMaterial(
    AbstractTestAbstractBackilluminatedCCDMaterial,
):
    def test_quantum_efficiency_measured(
        self,
        a: optika.sensors.AbstractStern1994BackilluminatedCCDMaterial,
    ):
        result = a.quantum_efficiency_measured
        assert isinstance(result, na.AbstractFunctionArray)
        assert np.all(result.outputs >= 0)
        assert np.all(result.outputs <= 1)
        assert np.all(result.inputs >= 0 * u.nm)



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from . import test_mixins


_position = [
    na.Cartesian3dVectorArray() * u.mm,
    na.Cartesian3dVectorArray(
        x=na.linspace(-5, 5, axis="x", num=3) * u.mm,
        y=na.linspace(-5, 5, axis="y", num=4) * u.mm,
        z=0 * u.mm,
    ),
]


class AbstractTestAbstractRulings(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestTransformable,
):
    @pytest.mark.parametrize("position", _position)
    def test_spacing(
        self,
        a: optika.rulings.AbstractRulings,
        position: na.AbstractCartesian3dVectorArray,
    ):
        result = a.spacing(position)
        assert isinstance(na.as_named_array(result), na.AbstractScalar)
        assert np.all(result > 0)
        assert na.unit_normalized(result).is_equivalent(u.mm)

    def test_diffraction_order(self, a: optika.rulings.AbstractRulings):
        assert na.get_dtype(a.diffraction_order) == int

    @pytest.mark.parametrize("position", _position)
    def test_normal(
        self,
        a: optika.rulings.AbstractRulings,
        position: na.AbstractCartesian3dVectorArray,
    ):
        result = a.normal(position)
        assert isinstance(result, na.AbstractCartesian3dVectorArray)
        assert np.all(result.length == 1)


class AbstractTestAbstractPolynomialDensityRulings(
    AbstractTestAbstractRulings,
):
    def test_coefficients(
        self,
        a: optika.rulings.AbstractPolynomialDensityRulings,
    ):
        result = a.coefficients
        for power, coefficient in result.items():
            assert isinstance(power, int)
            assert isinstance(na.as_named_array(coefficient), na.AbstractScalar)
            assert na.unit_normalized(coefficient).is_equivalent(u.mm ** -(power + 1))

    @pytest.mark.parametrize("position", _position)
    def test_frequency(
        self,
        a: optika.rulings.AbstractPolynomialDensityRulings,
        position: na.AbstractCartesian3dVectorArray,
    ):
        result = a.frequency(position)
        assert isinstance(na.as_named_array(result), na.AbstractScalar)
        assert np.all(result > 0)
        assert na.unit_normalized(result).is_equivalent(1 / u.mm)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.PolynomialDensityRulings(
            coefficients={
                0: 5000 / (u.um**0) / u.mm,
                1: 5 / (u.um**1) / u.mm,
                2: 6 / (u.um**2) / u.mm,
            },
            diffraction_order=1,
        ),
    ],
)
class TestPolynomialDensityRulings(
    AbstractTestAbstractPolynomialDensityRulings,
):
    pass


class AbstractTestAbstractConstantDensityRulings(
    AbstractTestAbstractPolynomialDensityRulings,
):
    def test_density(
        self,
        a: optika.rulings.AbstractConstantDensityRulings,
    ):
        assert a.density.unit.is_equivalent(1 / u.mm)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.ConstantDensityRulings(
            density=5000 / u.mm,
            diffraction_order=1,
        ),
        optika.rulings.ConstantDensityRulings(
            density=na.linspace(1, 5, axis="rulings", num=4) / u.mm,
            diffraction_order=na.ScalarArray(np.array([-1, 0, 1]), axes="m"),
        ),
    ],
)
class TestConstantDensityRulings(
    AbstractTestAbstractConstantDensityRulings,
):
    pass


class AbstractTestAbstractPolynomialSpacingRulings(
    AbstractTestAbstractRulings,
):
    def test_coefficients(
        self,
        a: optika.rulings.AbstractPolynomialSpacingRulings,
    ):
        result = a.coefficients
        for power, coefficient in result.items():
            assert isinstance(power, int)
            assert isinstance(na.as_named_array(coefficient), na.AbstractScalar)
            assert na.unit_normalized(coefficient).is_equivalent(u.mm ** -(power - 1))


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.PolynomialSpacingRulings(
            coefficients={
                0: 5 * u.um / (u.um**0),
                1: 5 * u.um / (u.um**1),
                2: 6 * u.um / (u.um**2),
            },
            diffraction_order=1,
        ),
    ],
)
class TestPolynomialSpacingRulings(
    AbstractTestAbstractPolynomialSpacingRulings,
):
    pass


class AbstractTestAbstractConstantSpacingRulings(
    AbstractTestAbstractPolynomialSpacingRulings,
):
    def test_period(
        self,
        a: optika.rulings.AbstractConstantSpacingRulings,
    ):
        assert a.period.unit.is_equivalent(u.mm)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.ConstantSpacingRulings(
            period=5 * u.um,
            diffraction_order=1,
        ),
        optika.rulings.ConstantSpacingRulings(
            period=na.linspace(1, 5, axis="rulings", num=4) * u.um,
            diffraction_order=na.ScalarArray(np.array([-1, 0, 1]), axes="m"),
        ),
    ],
)
class TestConstantSpacingRulings(
    AbstractTestAbstractConstantSpacingRulings,
):
    pass



import abc
import dataclasses
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractRulings",
    "AbstractPolynomialDensityRulings",
    "PolynomialDensityRulings",
    "AbstractConstantDensityRulings",
    "ConstantDensityRulings",
    "AbstractPolynomialSpacingRulings",
    "PolynomialSpacingRulings",
    "AbstractConstantSpacingRulings",
    "ConstantSpacingRulings",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRulings(
    optika.mixins.Printable,
    optika.mixins.Transformable,
):
    """
    Interface for the interaction of a ruled surface with incident light
    """

    @property
    @abc.abstractmethod
    def diffraction_order(self) -> na.ScalarLike:
        """
        the diffraction order to simulate
        """

    @abc.abstractmethod
    def spacing(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        """
        spacing between adjacent rulings at the given position

        Parameters
        ----------
        position
            location to evaluate the ruling spacing
        """

    @abc.abstractmethod
    def normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        """
        vector normal to the planes of the rulings

        Parameters
        ----------
        position
            location to evaluate the normal vector
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPolynomialDensityRulings(
    AbstractRulings,
):
    @property
    @abc.abstractmethod
    def coefficients(self) -> None | dict[int, na.ScalarLike]:
        """The coefficients of the polynomial describing the ruling density"""

    def frequency(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        """
        Density of rulings at the given position.
        Equivalent to the multiplicative inverse of the ruling spacing.

        Parameters
        ----------
        position
            location to evaluate the ruling density
        """
        transformation = self.transformation
        if transformation is not None:
            position = transformation(position)

        x = position @ self.normal(position)

        coefficients = self.coefficients
        if coefficients is None:
            coefficients = dict()

        result = 0 / u.mm
        for power, coefficient in coefficients.items():
            result = result + coefficient * (x**power)

        return result

    def spacing(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        return 1 / self.frequency(position)

    def normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        return na.Cartesian3dVectorArray(1, 0, 0)


@dataclasses.dataclass(eq=False, repr=False)
class PolynomialDensityRulings(
    AbstractPolynomialDensityRulings,
):
    coefficients: None | dict[int, na.ScalarLike] = None
    diffraction_order: na.ScalarLike = 1
    transformation: None | na.transformations.AbstractTransformation = None


@dataclasses.dataclass(eq=False, repr=False)
class AbstractConstantDensityRulings(
    AbstractPolynomialDensityRulings,
):
    @property
    @abc.abstractmethod
    def density(self) -> na.ScalarLike:
        """
        the frequency of the ruling pattern
        """

    @property
    def coefficients(self) -> None | dict[int, na.ScalarLike]:
        return {0: self.density}


@dataclasses.dataclass(eq=False, repr=False)
class ConstantDensityRulings(
    AbstractConstantDensityRulings,
):
    density: na.ScalarLike = 0 / u.mm
    diffraction_order: na.ScalarLike = 1
    transformation: None | na.transformations.AbstractTransformation = None


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPolynomialSpacingRulings(
    AbstractRulings,
):
    @property
    @abc.abstractmethod
    def coefficients(self) -> None | dict[int, na.ScalarLike]:
        """The coefficients of the polynomial describing the ruling density"""

    def spacing(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        transformation = self.transformation
        if transformation is not None:
            position = transformation(position)

        x = position @ self.normal(position)

        coefficients = self.coefficients
        if coefficients is None:
            coefficients = dict()

        result = 0 * u.mm
        for power, coefficient in coefficients.items():
            result = result + coefficient * (x**power)

        return result

    def normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        return na.Cartesian3dVectorArray(1, 0, 0)


@dataclasses.dataclass(eq=False, repr=False)
class PolynomialSpacingRulings(
    AbstractPolynomialSpacingRulings,
):
    coefficients: None | dict[int, na.ScalarLike] = None
    diffraction_order: na.ScalarLike = 1
    transformation: None | na.transformations.AbstractTransformation = None


@dataclasses.dataclass(eq=False, repr=False)
class AbstractConstantSpacingRulings(
    AbstractPolynomialSpacingRulings,
):
    @property
    @abc.abstractmethod
    def period(self) -> na.ScalarLike:
        """
        the spacing of the ruling pattern
        """

    @property
    def coefficients(self) -> None | dict[int, na.ScalarLike]:
        return {0: self.period}


@dataclasses.dataclass(eq=False, repr=False)
class ConstantSpacingRulings(
    AbstractConstantSpacingRulings,
):
    period: na.ScalarLike = 0 * u.mm
    diffraction_order: na.ScalarLike = 1
    transformation: None | na.transformations.AbstractTransformation = None



import abc
import functools
import dataclasses
import numpy as np
import scipy.optimize
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "energy_bandgap",
    "energy_electron_hole",
    "quantum_yield_ideal",
    "quantum_efficiency_effective",
    "AbstractImagingSensorMaterial",
    "AbstractCCDMaterial",
    "AbstractBackilluminatedCCDMaterial",
    "AbstractStern1994BackilluminatedCCDMaterial",
]

energy_bandgap = 1.12 * u.eV
"""the bandgap energy of silicon"""

energy_electron_hole = 3.65 * u.eV
"""
the high-energy limit of the energy required to create an electron-hole pair
in silicon at room temperature
"""


def quantum_yield_ideal(
    wavelength: u.Quantity | na.AbstractScalar,
) -> na.AbstractScalar:
    r"""
    Calculate the ideal quantum yield of a silicon detector for a given
    wavelength.

    Parameters
    ----------
    wavelength
        the wavelength of the incident photons

    Notes
    -----
    The quantum yield is the number of electron-hole pairs produced per photon.

    The ideal quantum yield is given in :cite:t:`Janesick2001` as:

    .. math::

        \text{QY}(\epsilon) = \begin{cases}
            0, & \epsilon < E_\text{g}\\
            1, &  E_\text{g} \leq \epsilon < E_\text{e-h} \\
            E_\text{e-h} / \epsilon, & E_\text{e-h} \leq \epsilon,
        \end{cases},

    where :math:`\epsilon` is the energy of the incident photon,
    :math:`E_\text{g} = 1.12\;\text{eV}` is the bandgap energy of silicon,
    and :math:`E_\text{e-h} = 3.65\;\text{eV}` is the energy required to
    generate 1 electron-hole pair in silicon at room temperature.

    Examples
    --------

    Plot the quantum yield vs wavelength

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of wavelengths
        wavelength = na.geomspace(100, 100000, axis="wavelength", num=101) << u.AA

        # Compute the quantum yield
        qy = optika.sensors.quantum_yield_ideal(wavelength)

        # Plot the quantum yield vs wavelength
        fig, ax = plt.subplots()
        na.plt.plot(wavelength, qy, ax=ax);
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("quantum yield");
    """
    energy = wavelength.to(u.eV, equivalencies=u.spectral())

    result = energy / energy_electron_hole
    result = np.where(energy > energy_electron_hole, result, 1)
    result = np.where(energy > energy_bandgap, result, 0)

    return result


def quantum_efficiency_effective(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: None | na.AbstractCartesian3dVectorArray = None,
    n: float | na.AbstractScalar = 1,
    thickness_oxide: u.Quantity | na.AbstractScalar = 50 * u.AA,
    thickness_implant: u.Quantity | na.AbstractScalar = 2317 * u.AA,
    thickness_substrate: u.Quantity | na.AbstractScalar = 7 * u.um,
    cce_backsurface: u.Quantity | na.AbstractScalar = 0.21,
    chemical_oxide: str | optika.chemicals.AbstractChemical = "SiO2",
    chemical_substrate: str | optika.chemicals.AbstractChemical = "Si",
    normal: None | na.AbstractCartesian3dVectorArray = None,
) -> na.AbstractScalar:
    r"""
    Calculate the effective quantum efficiency of a back-illuminated detector.

    Parameters
    ----------
    wavelength
        The wavelength of the incident light in vacuum.
    direction
        The propagation direction of the incident light in the ambient medium.
        If :obj:`None` (default), normal incidence (:math:`\hat{z}`) is assumed.
    n
        The complex index of refraction of the ambient medium.
    thickness_oxide
        The thickness of the oxide layer on the back surface of the sensor.
        Default is the value given in :cite:t:`Stearn1994`.
    thickness_implant
        The thickness of the implant layer.
        Default is the value given in :cite:t:`Stearn1994`.
    thickness_substrate
        The thickness of the silicon substrate.
        Default is the value given in :cite:t:`Stearn1994`.
    cce_backsurface
        The differential charge collection efficiency on the back surface
        of the sensor.
        Default is the value given in :cite:t:`Stearn1994`.
    chemical_oxide
        The chemical composition of the oxide layer.
        The default is to assume the oxide layer is silicon dioxide.
    chemical_substrate
        Optional complex refractive index of the implant region and substrate.
        The default is to assume the substrate is made from silicon.
    normal
        The vector perpendicular to the surface of the sensor.
        If :obj:`None`, then the normal is assumed to be :math:`-\hat{z}`

    Examples
    --------
    Reproduce Figure 12 from :cite:t:`Stern1994`, the modeled quantum efficiency
    of a Tektronix TK512CB :math:`512 \times 512` pixel backilluminated CCD.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import numpy as np
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of wavelengths with which to sample the EQE
        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Compute the effective quantum efficiency
        eqe = optika.sensors.quantum_efficiency_effective(
            wavelength=wavelength,
        )

        # Compute the maximum theoretical quantum efficiency
        eqe_max = optika.sensors.quantum_efficiency_effective(
            wavelength=wavelength,
            cce_backsurface=1,
        )

        # Plot the effective and maximum quantum efficiency
        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.plot(
            wavelength,
            eqe,
            ax=ax,
            label="effective quantum efficiency",
        );
        na.plt.plot(
            wavelength,
            eqe_max,
            ax=ax,
            label="maximum quantum efficiency",
        );
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("efficiency");
        ax.legend();

    Plot the EQE as a function of wavelength for normal and oblique incidence

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import numpy as np
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of wavelengths with which to sample the EQE
        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Define the incidence directions
        angle = na.linspace(0, 30, axis="angle", num=2) * u.deg
        direction = na.Cartesian3dVectorArray(
            x=np.sin(angle),
            y=0,
            z=np.cos(angle),
        )

        eqe = optika.sensors.quantum_efficiency_effective(
            wavelength=wavelength,
            direction=direction,
        )

        # Plot the results
        fig, ax = plt.subplots(constrained_layout=True)
        angle_str = angle.value.astype(str).astype(object)
        na.plt.plot(
            wavelength,
            eqe,
            ax=ax,
            axis="wavelength",
            label=r"$\theta$ = " + angle_str + f"{angle.unit:latex_inline}",
        );
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("efficiency");
        ax.legend();

    Notes
    -----
    Our goal is to recover Equation 11 in :cite:t:`Stern1994`, along wtih the
    correction for photons lost by transmission through the entire CCD substrate.
    From inspecting Equations 6 and 9 in :cite:t:`Stern1994`,
    we can see that the effective quantum efficiency is:

    .. math::
        :label: eqe-definition

        \text{EQE} = T_\lambda \int_0^\infty \alpha \eta(x) e^{-\alpha x} \; dx

    where :math:`T_\lambda` is the net transmission of photons through the backsurface
    oxide layer (accounting for both absorption and reflections) calculated using
    :func:`optika.materials.multilayer_efficiency`,
    :math:`\alpha` is the absorption coefficient of silicon,
    :math:`x` is the distance from the backsurface,
    and :math:`\eta(x)` is the differential charge collection efficiency (CCE).

    :cite:t:`Stern1994` assumes that the differential CCE takes the following
    linear form,

    .. math::
        :label: differential-cce

        \eta(x) = \begin{cases}
            \eta_0 + (1 - \eta_0) x / W, & 0 < x < W \\
            1, & W < x < D, \\
            0, & D < x
        \end{cases}

    where :math:`\eta_0` is the differential CCE at the backsurface,
    :math:`W` is the thickness of the implant region,
    and :math:`D` is the total thickness of the silicon substrate.

    Plugging Equation :eq:`differential-cce` into Equation :eq:`eqe-definition`
    and integrating yields

    .. math::
        :label: eqe

        \text{EQE} &= \alpha T_\lambda \left\{
                        \int_0^W \left[ \eta_0 + \left( \frac{1 - \eta_0}{W} \right) x \right] e^{-\alpha x} \; dx
                        + \int_W^D e^{-\alpha x} \; dx
        \right\} \\
        &= \alpha T_\lambda \left\{
            \eta_0 \int_0^W e^{-\alpha x} \; dx
            + \left( \frac{1 - \eta_0}{W} \right) \int_0^W x e^{-\alpha x} \; dx
            + \int_W^D e^{-\alpha x} \; dx
        \right\} \\
        &= \alpha T_\lambda \left\{
            -\left[ \frac{\eta_0}{\alpha} e^{-\alpha x} \right|_0^W
            - \left( \frac{1 - \eta_0}{W} \right) \left[ \left( \frac{\alpha x + 1}{\alpha^2} \right) e^{-\alpha x} \right|_0^W
            - \left[ \frac{1}{\alpha} e^{-\alpha x} \right|_W^D
        \right\} \\
        &= T_\lambda \left\{
            - \left[ \eta_0 (e^{-\alpha W} - 1) \right]
            - \left( \frac{1 - \eta_0}{\alpha W} \right) \left[ (\alpha W + 1) e^{-\alpha W} - 1 \right]
            - \left[ e^{-\alpha D} - e^{-\alpha W} \right]
        \right\} \\
        &= T_\lambda \left\{
            - \eta_0 e^{-\alpha W}
            + \eta_0
            - e^{-\alpha W}
            + \eta_0 e^{-\alpha W}
            + \left( \frac{1 - \eta_0}{\alpha W} \right) (1 - e^{-\alpha W})
            - e^{-\alpha D}
            + e^{-\alpha W}
        \right\} \\
        &= T_\lambda \left\{
            \eta_0
            + \left( \frac{1 - \eta_0}{\alpha W} \right) (1 - e^{-\alpha W})
            - e^{-\alpha D}
        \right\} \\

    Equation :eq:`eqe` is equivalent to Equation 11 in :cite:t:`Stern1994`,
    with the addition of an :math:`e^{-\alpha W}-e^{-\alpha D}` term which represents photons
    that traveled all the way through the silicon substrate without interacting.

    Equation :eq:`eqe` is only valid for normally-incident light.
    We can generalize it to obliquely-incident light by making the substitution

    .. math::
        :label: x-oblique

        x \rightarrow \frac{x}{\cos \theta}

    where :math:`\theta` is the angle between the propagation direction
    inside the silicon substrate and the normal vector.

    Substituting :eq:`x-oblique` into Equation :eq:`eqe` and solving yields

    .. math::
        :label: eqe-oblique

        \text{EQE} = T_\lambda \left\{
            \eta_0
            + \left( \frac{1 - \eta_0}{\alpha W \sec \theta} \right) (1 - e^{-\alpha W \sec \theta})
            - e^{-\alpha D \sec \theta}
        \right\} \\

    """
    if direction is None:
        direction = na.Cartesian3dVectorArray(0, 0, 1)

    if not isinstance(chemical_oxide, optika.chemicals.AbstractChemical):
        chemical_oxide = optika.chemicals.Chemical(chemical_oxide)

    if not isinstance(chemical_substrate, optika.chemicals.AbstractChemical):
        chemical_substrate = optika.chemicals.Chemical(chemical_substrate)

    if normal is None:
        normal = na.Cartesian3dVectorArray(0, 0, -1)

    reflectivity, transmissivity = optika.materials.multilayer_efficiency(
        wavelength=wavelength,
        direction=-direction @ normal,
        n=n,
        layers=optika.materials.Layer(
            chemical=chemical_oxide,
            thickness=thickness_oxide,
        ),
        substrate=optika.materials.Layer(
            chemical=chemical_substrate,
        ),
    )

    n_substrate = chemical_substrate.n(wavelength)
    wavenumber_substrate = np.imag(n_substrate)
    absorption_substrate = 4 * np.pi * wavenumber_substrate / wavelength

    direction_substrate = optika.materials.snells_law(
        wavelength=wavelength,
        direction=direction,
        index_refraction=np.real(n),
        index_refraction_new=np.real(n_substrate),
        normal=normal,
    )

    cos_theta = -direction_substrate @ normal

    z0 = absorption_substrate * thickness_implant / cos_theta
    exp_z0 = np.exp(-z0)

    term_1 = cce_backsurface
    term_2 = ((1 - cce_backsurface) / z0) * (1 - exp_z0)
    term_3 = -np.exp(-absorption_substrate * thickness_substrate / cos_theta)

    result = transmissivity.average * (term_1 + term_2 + term_3)

    return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImagingSensorMaterial(
    optika.materials.AbstractMaterial,
):
    """
    An interface representing the light-sensitive material of an imaging sensor.
    """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCCDMaterial(
    AbstractImagingSensorMaterial,
):
    """
    An interface representing the light-sensitive material of a CCD sensor.
    """

    @property
    def transformation(self) -> None:
        return None

    @functools.cached_property
    def _chemical(self) -> optika.chemicals.Chemical:
        return optika.chemicals.Chemical("Si")

    @functools.cached_property
    def _chemical_oxide(self) -> optika.chemicals.Chemical:
        return optika.chemicals.Chemical("SiO2")

    def index_refraction(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> na.ScalarLike:
        result = self._chemical.index_refraction(rays.wavelength)
        return result

    def attenuation(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> na.ScalarLike:
        result = self._chemical.wavenumber(rays.wavelength)
        result = 4 * np.pi * result / rays.wavelength
        return result

    @property
    def is_mirror(self) -> bool:
        return False


@dataclasses.dataclass(eq=False, repr=False)
class AbstractBackilluminatedCCDMaterial(
    AbstractCCDMaterial,
):
    """
    An interface representing the light-sensitive material of a backilluminated
    CCD sensor.
    """

    @property
    @abc.abstractmethod
    def thickness_oxide(self) -> u.Quantity | na.AbstractScalar:
        """
        The thickness of the oxide layer on the back surface of the CCD sensor.
        """

    @property
    @abc.abstractmethod
    def thickness_implant(self) -> u.Quantity | na.AbstractScalar:
        """
        The thickness of the implant layer of the CCD sensor.
        """

    @property
    @abc.abstractmethod
    def thickness_substrate(self) -> u.Quantity | na.AbstractScalar:
        """the thickness of the entire CCD silicon substrate"""

    @property
    @abc.abstractmethod
    def cce_backsurface(self) -> float | na.AbstractScalar:
        """
        The charge collection efficiency on the backsurface of the CCD sensor.
        """

    def quantum_yield_ideal(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> u.Quantity | na.AbstractScalar:
        """
        Compute the ideal quantum yield of this CCD sensor material using
        :func:`optika.sensors.quantum_yield_ideal`.

        Parameters
        ----------
        wavelength
            The wavelength of the incident light
        """
        return quantum_yield_ideal(wavelength)

    def quantum_efficiency_effective(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Compute the effective quantum efficiency of this CCD material using
        :func:`optika.sensors.quantum_efficiency_effective`.

        Parameters
        ----------
        rays
            The light rays incident on the CCD surface
        normal
            The vector perpendicular to the surface of the CCD.
        """
        k = rays.attenuation * rays.wavelength / (4 * np.pi)
        n = rays.index_refraction + k * 1j

        return quantum_efficiency_effective(
            wavelength=rays.wavelength,
            direction=rays.direction,
            n=n,
            thickness_oxide=self.thickness_oxide,
            thickness_implant=self.thickness_implant,
            thickness_substrate=self.thickness_substrate,
            cce_backsurface=self.cce_backsurface,
            chemical_oxide=self._chemical_oxide,
            chemical_substrate=self._chemical,
            normal=normal,
        )

    def efficiency(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        return self.quantum_efficiency_effective(
            rays=rays,
            normal=normal,
        )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractStern1994BackilluminatedCCDMaterial(
    AbstractBackilluminatedCCDMaterial,
):
    """
    A CCD material that is uses the method of :cite:t:`Stern1994` to compute
    the quantum efficiency.
    """

    @property
    @abc.abstractmethod
    def quantum_efficiency_measured(self) -> na.FunctionArray:
        """
        The measured quantum efficiency that will be fit by the function
        :func:`optika.sensors.quantum_efficiency_effective`.
        """

    @functools.cached_property
    def _quantum_efficiency_fit(self) -> dict[str, float | u.Quantity]:
        qe_measured = self.quantum_efficiency_measured

        unit_thickness_oxide = u.AA
        unit_thickness_implant = u.AA
        unit_thickness_substrate = u.um
        unit_cce_backsurface = u.dimensionless_unscaled

        def eqe_rms_difference(x: tuple[float, float, float, float]):
            (
                thickness_oxide,
                thickness_implant,
                thickness_substrate,
                cce_backsurface,
            ) = x
            qe_fit = quantum_efficiency_effective(
                wavelength=qe_measured.inputs,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
                thickness_oxide=thickness_oxide << unit_thickness_oxide,
                thickness_implant=thickness_implant << unit_thickness_implant,
                thickness_substrate=thickness_substrate << unit_thickness_substrate,
                cce_backsurface=cce_backsurface << unit_cce_backsurface,
            )

            return np.sqrt(np.mean(np.square(qe_measured.outputs - qe_fit))).ndarray

        thickness_oxide_guess = 50 * u.AA
        thickness_implant_guess = 2317 * u.AA
        thickness_substrate_guess = 7 * u.um
        cce_backsurface_guess = 0.21 * u.dimensionless_unscaled

        fit = scipy.optimize.minimize(
            fun=eqe_rms_difference,
            x0=[
                thickness_oxide_guess.to_value(unit_thickness_oxide),
                thickness_implant_guess.to_value(unit_thickness_implant),
                thickness_substrate_guess.to_value(unit_thickness_substrate),
                cce_backsurface_guess.to_value(unit_cce_backsurface),
            ],
            method="nelder-mead",
        )

        thickness_oxide, thickness_implant, thickness_substrate, cce_backsurface = fit.x
        thickness_oxide = thickness_oxide << unit_thickness_oxide
        thickness_implant = thickness_implant << unit_thickness_implant
        thickness_substrate = thickness_substrate << unit_thickness_substrate
        cce_backsurface = cce_backsurface << unit_cce_backsurface

        return dict(
            thickness_oxide=thickness_oxide,
            thickness_implant=thickness_implant,
            thickness_substrate=thickness_substrate,
            cce_backsurface=cce_backsurface,
        )

    @property
    def thickness_oxide(self) -> u.Quantity:
        return self._quantum_efficiency_fit["thickness_oxide"]

    @property
    def thickness_implant(self) -> u.Quantity:
        return self._quantum_efficiency_fit["thickness_implant"]

    @property
    def thickness_substrate(self) -> u.Quantity:
        return self._quantum_efficiency_fit["thickness_substrate"]

    @property
    def cce_backsurface(self) -> float:
        return self._quantum_efficiency_fit["cce_backsurface"]



import astropy.units as u
import named_arrays as na
from .._materials import AbstractStern1994BackilluminatedCCDMaterial

__all__ = [
    "TektronixTK512CBMaterial",
]


class TektronixTK512CBMaterial(
    AbstractStern1994BackilluminatedCCDMaterial,
):
    """
    A model of the light-sensitive material of a Tektronix TK512CB sensor based on
    measurements by :cite:t:`Stern1994`.

    Examples
    --------

    Plot the measured TK512CB quantum efficiency vs the fitted
    quantum efficiency calculated using the method of :cite:t:`Stern1994`.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Create a new instance of the e2v CCD97 light-sensitive material
        material_ccd = optika.sensors.TektronixTK512CBMaterial()

        # Store the wavelengths at which the QE was measured
        wavelength_measured = material_ccd.quantum_efficiency_measured.inputs

        # Store the QE measurements
        qe_measured = material_ccd.quantum_efficiency_measured.outputs

        # Define a grid of wavelengths with which to evaluate the fitted QE
        wavelength_fit = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Evaluate the fitted QE using the given wavelengths
        qe_fit = material_ccd.quantum_efficiency_effective(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength_fit,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the measured QE vs the fitted QE
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(constrained_layout=True)
            na.plt.scatter(
                wavelength_measured,
                qe_measured,
                label="measured",
            )
            na.plt.plot(
                wavelength_fit,
                qe_fit,
                label="fit",
            )
            ax.set_xscale("log")
            ax.set_xlabel(f"wavelength ({wavelength_fit.unit:latex_inline})")
            ax.set_ylabel("quantum efficiency")
            ax.legend()

    The thickness of the oxide layer found by the fit is

    .. jupyter-execute::

        material_ccd.thickness_oxide

    The thickness of the implant layer found by the fit is

    .. jupyter-execute::

        material_ccd.thickness_implant

    The thickness of the substrate found by the fit is

    .. jupyter-execute::

        material_ccd.thickness_substrate

    And the differential charge collection efficiency at the backsurface
    found by the fit is

    .. jupyter-execute::

        material_ccd.cce_backsurface
    """

    @property
    def quantum_efficiency_measured(self) -> na.FunctionArray:
        wavelength = [
            13.3,
            23.6,
            44.7,
            67.6,
            114.0,
            135.5,
            171.4,
            256.0,
            303.8,
            461.0,
            584.0,
            736.0,
            1215.5,
            2537.0,
            3650.0,
            4050.0,
        ] * u.AA
        qe = [
            0.91,
            0.80,
            0.48,
            0.32,
            0.42,
            0.86,
            0.82,
            0.60,
            0.58,
            0.53,
            0.30,
            0.085,
            0.055,
            0.06,
            0.09,
            0.29,
        ] * u.dimensionless_unscaled
        return na.FunctionArray(
            inputs=na.ScalarArray(wavelength, axes="wavelength"),
            outputs=na.ScalarArray(qe, axes="wavelength"),
        )



from __future__ import annotations
from typing import Sequence, TypeVar, Generic, Callable, Any
import abc
import dataclasses
import functools
import astropy.units as u
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import named_arrays as na
import optika

__all__ = [
    "AbstractSystem",
    "AbstractSequentialSystem",
    "SequentialSystem",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSystem(
    optika.mixins.Plottable,
    optika.mixins.Printable,
    optika.mixins.Transformable,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSequentialSystem(
    AbstractSystem,
):
    @property
    @abc.abstractmethod
    def object(self) -> None | optika.surfaces.AbstractSurface:
        """
        The external object being imaged or illuminated by this system.

        If :obj:`None`, the external object is assumed to be at infinity.
        """

    @property
    @abc.abstractmethod
    def surfaces(self) -> Sequence[optika.surfaces.AbstractSurface]:
        """
        a sequence of surfaces representing this optical system.

        At least one of these surfaces needs to be marked as the pupil surface,
        and if the object surface is not marked as the field stop, one of these
        surfaces needs to be marked as the field stop.
        """

    @property
    @abc.abstractmethod
    def axis_surface(self) -> str:
        """
        The name of the logical axis representing the sequence of surfaces.
        """

    @property
    def surfaces_all(self) -> list[optika.surfaces.AbstractSurface]:
        """
        Concatenate :attr:`object` with :attr:`surfaces` into a single list of
        surfaces.
        """
        obj = self.object
        if obj is not None:
            result = [obj]
        else:
            result = []
        result += list(self.surfaces)
        if not any(s.is_field_stop for s in result):
            result[0] = dataclasses.replace(result[0], is_field_stop=True)
        return result

    @property
    @abc.abstractmethod
    def grid_input_normalized(self) -> optika.vectors.ObjectVectorArray:
        """
        The input grid to sample with rays.

        This grid is projected onto the 3 possible types of stops: field, pupil,
        and spectral.

        Positions on the stop can be specified in either absolute or normalized
        coordinates. Using normalized coordinates allows for injecting different
        grid types (cylindrical, stratified random, etc.) without specifying the
        scale of the stop surface.

        If positions are specified in absolute units, they are measured in the
        coordinate system of the corresponding stop surface
        """

    @property
    def _indices_field_stop(self) -> list[int]:
        return [i for i, s in enumerate(self.surfaces_all) if s.is_field_stop]

    @property
    def _indices_pupil_stop(self) -> list[int]:
        return [i for i, s in enumerate(self.surfaces_all) if s.is_pupil_stop]

    @property
    def index_field_stop(self) -> int:
        indices = self._indices_field_stop
        if not indices:
            raise ValueError(
                "Field stop is not defined for this system."
                "Set `is_field_stop=True` for at least one surface in this system."
            )
        return indices[~0]

    @property
    def index_pupil_stop(self) -> int:
        indices = self._indices_pupil_stop
        if not indices:
            raise ValueError(
                "Pupil stop is not defined for this system."
                "Set `is_pupil_stop=True` for at least one surface in this "
                "system."
            )
        return indices[~0]

    @property
    def field_stop(self) -> optika.surfaces.AbstractSurface:
        return self.surfaces_all[self.index_field_stop]

    @property
    def pupil_stop(self) -> optika.surfaces.AbstractSurface:
        return self.surfaces_all[self.index_pupil_stop]

    @classmethod
    def _ray_error(
        cls,
        a: na.Cartesian2dVectorArray,
        rays: optika.rays.RayVectorArray,
        subsystem: list[optika.surfaces.AbstractSurface],
        grid_last: na.Cartesian2dVectorArray,
        component_variable: str,
        component_target: str,
        zfunc: Callable[[na.Cartesian2dVectorArray], na.ScalarLike],
    ):
        rays_component_variable = getattr(rays, component_variable)
        rays_component_variable.x = a.x
        rays_component_variable.y = a.y
        rays_component_variable.z = zfunc(a)

        rays = optika.propagators.propagate_rays(
            propagators=subsystem[1:],
            rays=rays,
        )

        transformation_last = subsystem[~0].transformation
        if transformation_last is not None:
            rays = transformation_last.inverse(rays)

        rays_component_target = getattr(rays, component_target)
        grid_last_trial = na.Cartesian2dVectorArray(
            x=rays_component_target.x,
            y=rays_component_target.y,
        )
        result = grid_last_trial - grid_last
        return result

    def _calc_rayfunction_input_stops_only(
        self,
        wavelength_input: na.ScalarLike,
        axis_pupil_stop: str,
        axis_field_stop: str,
        samples_pupil_stop: int = 101,
        samples_field_stop: int = 101,
    ) -> optika.rays.RayFunctionArray:
        result = optika.rays.RayFunctionArray(
            inputs=optika.vectors.ObjectVectorArray(
                wavelength=wavelength_input,
            ),
            outputs=optika.rays.RayVectorArray(
                wavelength=wavelength_input,
            ),
        )

        surfaces = self.surfaces_all

        indices_pupil_stop = self._indices_pupil_stop
        indices_field_stop = self._indices_field_stop

        if not indices_pupil_stop:
            raise ValueError("pupil not defined")
        if not indices_field_stop:
            raise ValueError("field stop not defined")

        while indices_pupil_stop and indices_field_stop:
            index_pupil_stop = indices_pupil_stop[0]
            index_field_stop = indices_field_stop[0]

            index_first = min(index_pupil_stop, index_field_stop)
            index_last = max(index_pupil_stop, index_field_stop)

            subsystem = surfaces[index_first : index_last + 1]

            surface_first = subsystem[0]
            surface_last = subsystem[~0]

            aperture_first = surface_first.aperture
            aperture_last = surface_last.aperture

            grid_first = np.moveaxis(
                a=aperture_first.wire(num=samples_pupil_stop),
                source="wire",
                destination=axis_pupil_stop,
            )
            grid_first = na.Cartesian2dVectorArray(grid_first.x, grid_first.y)

            grid_last = np.moveaxis(
                a=aperture_last.wire(num=samples_field_stop),
                source="wire",
                destination=axis_field_stop,
            )
            grid_last = na.Cartesian2dVectorArray(grid_last.x, grid_last.y)

            if surface_first.is_pupil_stop:
                indices_pupil_stop.pop(0)
                result.inputs.pupil = grid_first
                result.inputs.field = grid_last
            else:
                indices_field_stop.pop(0)
                result.inputs.field = grid_first
                result.inputs.pupil = grid_last

            if na.unit(grid_first).is_equivalent(u.mm):
                result.outputs.position = na.Cartesian3dVectorArray(
                    x=grid_first.x,
                    y=grid_first.y,
                    z=surface_first.sag(grid_first),
                )
                result.outputs.direction = na.Cartesian3dVectorArray(0, 0, 1)
                component_variable = "direction"
                zfunc = lambda xy: np.sqrt(1 - np.square(xy.length))
            elif na.unit(grid_first).is_equivalent(u.dimensionless_unscaled):
                result.outputs.direction = na.Cartesian3dVectorArray(
                    x=grid_first.x,
                    y=grid_first.y,
                    z=np.sqrt(1 - np.square(grid_first.length)),
                )
                result.outputs.position = na.Cartesian3dVectorArray() * u.mm
                component_variable = "position"
                zfunc = lambda xy: surface_first.sag(xy)
            else:
                raise ValueError(f"unrecognized input grid unit, {na.unit(grid_first)}")

            if surface_first.transformation is not None:
                result.outputs = surface_first.transformation(result.outputs)

            if na.unit(grid_last).is_equivalent(u.mm):
                component_target = "position"
            elif na.unit(grid_last).is_equivalent(u.dimensionless_unscaled):
                component_target = "direction"
            else:
                raise ValueError(f"unrecognized output grid unit, {na.unit(grid_last)}")

            variables = getattr(result.outputs, component_variable)

            root = na.optimize.root_newton(
                function=functools.partial(
                    self._ray_error,
                    rays=result.outputs,
                    subsystem=subsystem,
                    grid_last=grid_last,
                    component_variable=component_variable,
                    component_target=component_target,
                    zfunc=zfunc,
                ),
                guess=na.Cartesian2dVectorArray(
                    x=variables.x,
                    y=variables.y,
                ),
            )

            variables.x = root.x
            variables.y = root.y
            variables.z = zfunc(root)

        return result

    def _calc_rayfunction_input_stops(
        self,
        wavelength_input: na.ScalarLike,
        axis_pupil_stop: str,
        axis_field_stop: str,
        samples_pupil_stop: int = 101,
        samples_field_stop: int = 101,
    ) -> optika.rays.RayFunctionArray:
        surfaces = self.surfaces_all

        index_pupil_stop = self.index_pupil_stop
        index_field_stop = self.index_field_stop

        index_stop = min(index_pupil_stop, index_field_stop)

        subsystem = surfaces[index_stop::-1]

        rays_stop = self._calc_rayfunction_input_stops_only(
            wavelength_input=wavelength_input,
            axis_pupil_stop=axis_pupil_stop,
            axis_field_stop=axis_field_stop,
            samples_pupil_stop=samples_pupil_stop,
            samples_field_stop=samples_field_stop,
        )

        result = rays_stop.copy_shallow()
        result.outputs = optika.propagators.propagate_rays(
            propagators=subsystem,
            rays=rays_stop.outputs,
        )

        if self.transformation is not None:
            result.outputs = self.transformation(result.outputs)

        return result

    def _calc_rayfunction_input(
        self,
        grid_input: optika.vectors.ObjectVectorArray,
    ) -> optika.rays.RayFunctionArray:
        rayfunction_inputs_stops = self._calc_rayfunction_input_stops(
            wavelength_input=grid_input.wavelength,
            axis_pupil_stop="wire_pupil",
            axis_field_stop="wire_field_stop",
            samples_pupil_stop=21,
            samples_field_stop=21,
        )

        # return rayfunction_inputs_stops

        obj = self.object
        if obj is not None:
            aperture_obj = obj.aperture
            if aperture_obj is not None:
                unit_obj = na.unit_normalized(aperture_obj.bound_lower)
                if unit_obj.is_equivalent(u.mm):
                    object_at_infinity = False
                else:
                    object_at_infinity = True
            else:
                object_at_infinity = False
        else:
            object_at_infinity = True

        result = rayfunction_inputs_stops.copy_shallow()

        if object_at_infinity:
            field = rayfunction_inputs_stops.outputs.direction.xy
            pupil = rayfunction_inputs_stops.outputs.position.xy
        else:
            field = rayfunction_inputs_stops.outputs.position.xy
            pupil = rayfunction_inputs_stops.outputs.direction.xy

        axis_pupil = tuple(result.inputs.pupil.shape)
        axis_field = tuple(result.inputs.field.shape)

        field = (field.min(axis_pupil) + field.max(axis_pupil)) / 2
        pupil = (pupil.min(axis_field) + pupil.max(axis_field)) / 2

        min_field = field.min(axis=axis_field)
        min_pupil = pupil.min(axis=axis_pupil)

        ptp_field = field.ptp(axis=axis_field)
        ptp_pupil = pupil.ptp(axis=axis_pupil)

        result.inputs = grid_input.copy_shallow()
        result.inputs.field = ptp_field * (result.inputs.field + 1) / 2 + min_field
        result.inputs.pupil = ptp_pupil * (result.inputs.pupil + 1) / 2 + min_pupil

        if object_at_infinity:
            position = result.inputs.pupil
            direction = result.inputs.field
        else:
            position = result.inputs.field
            direction = result.inputs.pupil

        rays = optika.rays.RayVectorArray(
            wavelength=grid_input.wavelength,
            position=na.Cartesian3dVectorArray(
                x=position.x,
                y=position.y,
                z=0 * position.x.unit,
            ),
            direction=na.Cartesian3dVectorArray(
                x=direction.x,
                y=direction.y,
                z=np.sqrt(1 - np.square(direction.length)),
            ),
        )

        if obj is not None:
            if obj.transformation is not None:
                rays = obj.transformation(rays)

        result.outputs = rays

        return result

    @functools.cached_property
    def _rayfunction_input(self) -> optika.rays.RayFunctionArray:
        return self._calc_rayfunction_input(
            grid_input=self.grid_input_normalized,
        )

    def _calc_raytrace(
        self,
        rayfunction_input: optika.rays.RayFunctionArray,
        axis: str,
    ) -> optika.rays.RayFunctionArray:
        rays = rayfunction_input.outputs
        if self.transformation is not None:
            rays = self.transformation.inverse(rays)
        result = rayfunction_input.copy_shallow()
        result.outputs = optika.propagators.accumulate_rays(
            propagators=self.surfaces_all,
            rays=rays,
            axis=axis,
        )
        return result

    @property
    def raytrace(self) -> optika.rays.RayFunctionArray:
        return self._calc_raytrace(
            rayfunction_input=self._rayfunction_input,
            axis=self.axis_surface,
        )

    def _calc_rayfunction(
        self,
        rayfunction_input: optika.rays.RayFunctionArray,
        axis: str,
    ) -> optika.rays.RayFunctionArray:
        raytrace = self._calc_raytrace(
            rayfunction_input=rayfunction_input,
            axis=axis,
        )
        return raytrace[{axis: ~0}]

    @functools.cached_property
    def rayfunction(self) -> optika.rays.RayFunctionArray:
        return self._calc_rayfunction(
            rayfunction_input=self._rayfunction_input,
            axis=self.axis_surface,
        )

    def plot(
        self,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray] = None,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        plot_rays: bool = True,
        plot_rays_vignetted: bool = False,
        kwargs_rays: None | dict[str, Any] = None,
        **kwargs,
    ) -> na.AbstractScalar | dict[str, na.AbstractScalar]:
        surfaces = self.surfaces_all
        transformation_self = self.transformation
        kwargs_plot = self.kwargs_plot

        if transformation is not None:
            if transformation_self is not None:
                transformation = transformation @ transformation_self
        else:
            transformation = transformation_self

        if kwargs_plot is not None:
            kwargs = kwargs | kwargs_plot

        result = dict()

        result["surfaces"] = []
        for surface in surfaces:
            result["surfaces"].append(
                surface.plot(
                    ax=ax,
                    transformation=transformation,
                    components=components,
                    **kwargs,
                )
            )

        if plot_rays:
            raytrace = self.raytrace

            if kwargs_rays is None:
                kwargs_rays = dict()

            where = True
            if not plot_rays_vignetted:
                intensity = raytrace.outputs[dict(surface=~0)].intensity
                where = intensity != 0

            result["rays"] = na.plt.plot(
                raytrace.outputs.position,
                ax=ax,
                axis=self.axis_surface,
                where=where,
                transformation=transformation,
                components=components,
                **kwargs_rays,
            )

        return result


@dataclasses.dataclass(eq=False, repr=False)
class SequentialSystem(
    AbstractSequentialSystem,
):
    """
    A sequential optical system is a composition of a sequence of
    :class:`optika.surfaces.AbstractSurface` instances and a default
    grid to sample them with.

    Examples
    --------

    Here is an example of a simple Newtonian telescope, with a parabolic
    primary mirror, a 45 degree fold mirror, and a detector.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import matplotlib as mpl
        import numpy as np
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # store the z-coordinate of the primary  and fold mirror
        # so that we can determine the detector position
        primary_mirror_z = 300 * u.mm
        fold_mirror_z = 50 * u.mm

        # define the parabolic primary mirror
        primary_mirror = optika.surfaces.Surface(
            name="mirror",
            sag=optika.sags.ParabolicSag(-300 * u.mm),
            aperture=optika.apertures.RectangularAperture(40 * u.mm),
            material=optika.materials.Mirror(),
            is_pupil_stop=True,
            transformation=na.transformations.Cartesian3dTranslation(
                z=primary_mirror_z,
            ),
        )

        # define the flat fold mirror which directs light
        # to the detector surface
        fold_mirror = optika.surfaces.Surface(
            name="fold_mirror",
            aperture=optika.apertures.CircularAperture(15 * u.mm),
            material=optika.materials.Mirror(),
            transformation=na.transformations.TransformationList([
                na.transformations.Cartesian3dRotationX(45 * u.deg),
                na.transformations.Cartesian3dTranslation(
                    z=fold_mirror_z,
                ),
            ]),
        )

        # define the central obscuration, the back face
        # of the fold mirror which blocks some of the
        # incoming light
        obscuration = optika.surfaces.Surface(
            name="obscuration",
            aperture=optika.apertures.CircularAperture(
                radius=fold_mirror.aperture.radius,
                inverted=True,
            ),
            transformation=fold_mirror.transformation,
        )

        # define the detector surface
        detector = optika.surfaces.Surface(
            name="detector",
            aperture=optika.apertures.RectangularAperture(5 * u.mm),
            transformation=na.transformations.TransformationList([
                na.transformations.Cartesian3dRotationX(90 * u.deg),
                na.transformations.Cartesian3dTranslation(
                    y=primary_mirror.sag.focal_length + (primary_mirror_z - fold_mirror_z),
                    z=fold_mirror_z,
                ),
            ]),
            is_field_stop=True,
        )

        # define the grid of normalized field coordinates,
        # which are in the range +/-0.99 because the range +/-1
        # would be clipped by the edge of the detector
        field = 0.99 * na.Cartesian2dVectorLinearSpace(
            start=-1,
            stop=1,
            axis=na.Cartesian2dVectorArray("field_x", "field_y"),
            num=5,
        )

        # define the grid of normalized pupil coordinates
        # in a similar fashion to the normalized field
        # coordinates
        pupil = 0.99 * na.Cartesian2dVectorLinearSpace(
            start=-1,
            stop=1,
            axis=na.Cartesian2dVectorArray("pupil_x", "pupil_y"),
            num=5,
        )

        # define the optical system using the surfaces
        # and the normalized field/pupil coordinates
        system = optika.systems.SequentialSystem(
            surfaces=[
                obscuration,
                primary_mirror,
                fold_mirror,
                detector,
            ],
            grid_input_normalized=optika.vectors.ObjectVectorArray(
                wavelength=500 * u.nm,
                field=field,
                pupil=pupil,
            ),
        )

        # plot the system
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            ax.set_aspect("equal")
            system.plot(
                ax=ax,
                components=("z", "y"),
                kwargs_rays=dict(
                    color="tab:blue",
                ),
                color="black",
                zorder=10,
            )
    """

    surfaces: Sequence[optika.surfaces.AbstractSurface] = dataclasses.MISSING
    object: optika.surfaces.AbstractSurface = None
    axis_surface: str = "surface"
    grid_input_normalized: optika.vectors.ObjectVectorArray = None
    transformation: None | na.transformations.AbstractTransformation = None
    kwargs_plot: None | dict[str, Any] = None



from __future__ import annotations
from typing import TypeVar
from typing_extensions import Self
import dataclasses
import named_arrays as na
from . import AbstractPupilVectorArray, PupilVectorArray
from . import AbstractSceneVectorArray, SceneVectorArray


WavelengthT = TypeVar("WavelengthT", bound=na.ScalarLike)
FieldT = TypeVar("FieldT", bound=na.AbstractCartesian2dVectorArray)
PupilT = TypeVar("PupilT", bound=na.AbstractCartesian2dVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractObjectVectorArray(
    AbstractPupilVectorArray,
    AbstractSceneVectorArray,
):
    @property
    def type_abstract(self) -> type[AbstractObjectVectorArray]:
        return AbstractObjectVectorArray

    @property
    def type_explicit(self) -> type[ObjectVectorArray]:
        return ObjectVectorArray

    @property
    def type_matrix(self) -> type[na.AbstractMatrixArray]:
        raise NotImplementedError


@dataclasses.dataclass(eq=False, repr=False)
class ObjectVectorArray(
    AbstractObjectVectorArray,
    PupilVectorArray,
    SceneVectorArray,
):
    @classmethod
    def from_scalar(
        cls,
        scalar: na.AbstractScalar,
        like: None | Self = None,
    ) -> Self:
        if like is not None:
            return type(like)(
                wavelength=scalar,
                field=scalar,
                pupil=scalar,
            )
        else:
            return cls(
                wavelength=scalar,
                field=scalar,
                pupil=scalar,
            )



import pytest
import astropy.units as u
import named_arrays as na
import optika
from . import test_mixins


class AbstractTestAbstractSystem(
    test_mixins.AbstractTestPlottable,
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestTransformable,
):
    pass


class AbstractTestAbstractSequentialSystem(
    AbstractTestAbstractSystem,
):
    def test_object(self, a: optika.systems.AbstractSequentialSystem):
        if a.object is not None:
            assert isinstance(a.object, optika.surfaces.AbstractSurface)

    def test_surfaces(self, a: optika.systems.AbstractSequentialSystem):
        for surface in a.surfaces:
            assert isinstance(surface, optika.surfaces.AbstractSurface)

    def test_axis_surface(self, a: optika.systems.AbstractSequentialSystem):
        assert isinstance(a.axis_surface, str)

    def test_surfaces_all(self, a: optika.systems.AbstractSequentialSystem):
        for surface in a.surfaces_all:
            assert isinstance(surface, optika.surfaces.AbstractSurface)

    def grid_input_normalized(self, a: optika.systems.AbstractSequentialSystem):
        assert isinstance(a.grid_input_normalized, optika.vectors.ObjectVectorArray)

    def test_index_field_stop(self, a: optika.systems.AbstractSequentialSystem):
        assert isinstance(a.index_field_stop, int)
        assert a.surfaces_all[a.index_field_stop].is_field_stop

    def test_index_pupil_stop(self, a: optika.systems.AbstractSequentialSystem):
        assert isinstance(a.index_pupil_stop, int)
        assert a.surfaces_all[a.index_pupil_stop].is_pupil_stop

    def test_field_stop(self, a: optika.systems.AbstractSequentialSystem):
        assert a.field_stop.is_field_stop

    def test_pupil_stop(self, a: optika.systems.AbstractSequentialSystem):
        assert a.pupil_stop.is_pupil_stop

    def test_raytrace(self, a: optika.systems.AbstractSequentialSystem):
        raytrace = a.raytrace
        assert isinstance(raytrace, optika.rays.RayFunctionArray)
        assert isinstance(raytrace.inputs, optika.vectors.ObjectVectorArray)
        assert isinstance(raytrace.outputs, optika.rays.RayVectorArray)
        assert a.axis_surface in raytrace.shape

    def test_rayfunction(self, a: optika.systems.AbstractSequentialSystem):
        rayfunction = a.rayfunction
        assert isinstance(rayfunction, optika.rays.RayFunctionArray)
        assert isinstance(rayfunction.inputs, optika.vectors.ObjectVectorArray)
        assert isinstance(rayfunction.outputs, optika.rays.RayVectorArray)
        assert a.axis_surface not in rayfunction.shape


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.systems.SequentialSystem(
            surfaces=[
                optika.surfaces.Surface(
                    name="mirror",
                    sag=optika.sags.SphericalSag(-200 * u.mm),
                    material=optika.materials.Mirror(),
                    aperture=optika.apertures.CircularAperture(20 * u.mm),
                    is_pupil_stop=True,
                    transformation=na.transformations.Cartesian3dTranslation(
                        z=100 * u.mm
                    ),
                ),
                optika.surfaces.Surface(
                    name="detector",
                    aperture=optika.apertures.RectangularAperture(10 * u.mm),
                    is_field_stop=True,
                ),
            ],
            grid_input_normalized=optika.vectors.ObjectVectorArray(
                wavelength=500 * u.nm,
                field=na.Cartesian2dVectorLinearSpace(
                    start=0,
                    stop=1,
                    axis=na.Cartesian2dVectorArray("field_x", "field_y"),
                    num=5,
                ),
                pupil=na.Cartesian2dVectorLinearSpace(
                    start=0,
                    stop=1,
                    axis=na.Cartesian2dVectorArray("pupil_x", "pupil_y"),
                    num=5,
                ),
            ),
        )
    ],
)
class TestSequentialSystem(AbstractTestAbstractSequentialSystem):
    pass



"""
Models of light sensors that can be used in optical systems.
"""

from typing import TypeVar
import abc
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractImagingSensor",
    "AbstractCCD",
]


MaterialT = TypeVar("MaterialT", bound=optika.materials.AbstractMaterial)


class AbstractImagingSensor(
    optika.surfaces.AbstractSurface[
        None,
        MaterialT,
        optika.apertures.RectangularAperture,
        optika.apertures.RectangularAperture,
        None,
    ],
):
    @property
    def sag(self) -> None:
        return None

    @property
    def rulings(self) -> None:
        return None


class AbstractCCD(
    AbstractImagingSensor[MaterialT],
):
    pass



"""
A Python package for simulating and designing optical systems.
"""

from . import mixins
from . import vectors
from . import rays
from . import metrology
from . import sags
from . import chemicals
from . import materials
from . import apertures
from . import rulings
from . import propagators
from . import surfaces
from . import sensors
from . import systems
